# --8<--8<--8<--8<--
#
# Copyright (C) 2010 Smithsonian Astrophysical Observatory
#
# This file is part of CXC::PDL::FuncND
#
# CXC::PDL::FuncND is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# -->8-->8-->8-->8--

package CXC::PDL::FuncND;

use strict;
use warnings;

use PDL;
use PDL::Transform qw[];
use PDL::Options qw[ iparse ];

use Math::Trig qw[ pi ];

use base 'PDL::Exporter';

our @EXPORT_OK = qw[
		       mahalanobis
		       gaussND
		  ];

our %EXPORT_TAGS = (Func     => [@EXPORT_OK],
                   );

# bizarre spurious errors are being thrown by this policy
## no critic (ProhibitAccessOfPrivateData)

our $VERSION = '0.04';

=head1 NAME

CXC::PDL::FuncND - N dimensional version of functions


=head1 SYNOPSIS

    use CXC::PDL::FuncND;

=for author to fill in:
    Brief code example(s) here showing commonest usage(s).
    This section will be as far as many users bother reading
    so make it as educational and exeplary as possible.


=head1 DESCRIPTION

This module provides multi-dimensional versions of one-dimensional functions
available elsewhere in PDL.

=cut


=pod

=head2 gaussND

=begin usage

  $g = gaussND( [OPTIONAL TYPE], $nx, $ny, ..., \%options );
  $g = gaussND( $piddle, \%options );
  gaussND( inplace $piddle, \%options );

=end usage

=for ref

Calculate the sampled multi-dimensional Gaussian PDF calculated on an
N-dimensional grid or at a set of locations.

=over

=item * calculating the PDF on a grid

Either specify the output piddle dimensions explicitly,

  $g = gaussND( [ OPTIONAL TYPE], $nx, $ny, ..., \%options );

or specify a template piddle I<without> specifying the C<vectors>
option:

  $g = gaussND( $piddle, \%options );

By default B<gaussND> will calculate the PDF at the I<indices> of the
points in the input piddle.  These may be mapped to other values by
specifying a transform with the C<transform> option.  B<gaussND> is
inplace aware, and will use B<$piddle> as the output piddle if its
inplace flag is set.

=item * calculating the PDF at a set of locations

The input piddle should represent a set of vectors and should have a
shape of (N,m), where C<m> is the number of vectors in the set. The
C<vectors> option must also be set:

  $piddle = pdl( [2,1], [3,1], [4,2]  );
  $g = gaussND( $piddle, { vectors => 1 } );

The vectors may be transformed before use via the C<transform> option.

=back

The following options are available:

=over

=item center | centre

The center of the Gaussian.  If not specified it defaults to the
origin.

This may take one of the following values:

=over

=item A vector of shape (N).

The location of the center. This may be either a Perl arrayref or a
one dimensional piddle.  If the input coordinates are transformed,
this is in the I<transformed> space.

=item the string C<auto>

If the PDF is calculated on a grid, this will center the Gaussian on
the grid. It is an error to use this for explicit locations.

=back

=item transform

A PDL::Transform object to be applied to the input coordinates.

=item scale

The Gaussian scale. If the input coordinates are transformed
via the C<transform> option, the units of scale are those in the
I<transformed> space.  This may be specified as:

=over

=item a scalar (Perl or piddle)

This results in a symmetric Gaussian with the given scale along each
coordinate.

=item a vector of shape (N) (piddle or Perl arrayref)

This results in a Gaussian with the specified scales for each
coordinate.

=item the full covariance matrix (piddle of shape (N,N))

This results in a Gaussian with correlated scales. At present this
matrix is not verified to be a legitimate covariance matrix.

=back

=item theta (Perl scalar)

B<Only for 2D!> Calculates a Gaussian PDF rotated by the specified
angle (in radians).  The C<scale> parameter must have shape (2).

=item log (Boolean)

If true, return the logarithm of the PDF. Defaults to false.

=back

=cut


sub gaussND {

    # handle being called as a method or a function
    my $self = eval { ref $_[0] && $_[0]->isa( 'PDL' ) } ? shift @_ : 'PDL';

    # handle options.
    my $opt = 'HASH' eq ref $_[-1] ? pop( @_ ) : {};
    my %opt = iparse( { center    => undef,
			scale     => 1,
			transform => undef,
			vectors   => 0,
			log       => 0,
			theta     => undef,
		      }, $opt );

    my ( $vectors, $output, $ndims, $center );

    if ( $opt{vectors} )
    {
	barf( "first argument must be a piddle if vectors option is set\n" )
	  unless eval { $_[0]->isa( 'PDL' ) };

	$output = shift;
	$vectors = $self;

	barf( "wrong dimensions for input vectors; expected 2 , got ",
	      $vectors->ndims, "\n" )
	  unless $vectors->ndims == 2;

	( $ndims, my $m ) = $vectors->dims;

	$vectors = $opt{transform}->apply( $vectors )
	  if defined $opt{transform};

	if ( defined $opt{center} )
	{
	    barf( "cannot use center = 'auto' if vectors is set\n" )
	      if $opt{center} eq 'auto';
	    $center = PDL::Core::topdl( $opt{center} );
	}
    }

    else
    {
	$output = @_ ? $self->new_from_specification( @_ )
	             : $self->new_or_inplace;

	$ndims = $output->ndims;

	$vectors = $output->ndcoords->reshape( $ndims, $output->nelem );

	$vectors->inplace->apply( $opt{transform} )
	  if defined $opt{transform};

	if ( defined $opt{center} )
	{
	    if ( $opt{center} eq 'auto' )
	    {
		$center = ( pdl([$output->dims]) - 1 ) / 2;
		$center->inplace->apply( $opt{transform} )
		  if defined $opt{transform};
	    }
	    else
	    {
		$center =  PDL::Core::topdl( $opt{center} )
	    }
	}
    }

    # for 1D Gaussians, $center may be a 0D PDL; this causes grief;
    # promote it to a 1D
    $center = $center->dummy(0) if defined $center && $center->ndims == 0;

    barf( "center vector has wrong dimensions\n" )
      if defined $center && ($center->ndims != 1
			     || ($center->dims)[0] != $ndims);

    my $cov;

    # special case; angle specification for 2D rotated uncorrelated Gaussian
    if ( defined $opt{theta} )
    {
	barf( "theta may only be used for 2D Gaussian PDFs\n" )
	  if $ndims != 2;

	my @s;
	# 1D piddle of length N
	if ( 'ARRAY' eq ref $opt{scale} && @{$opt{scale}} == $ndims)
	{
	    @s = @{$opt{scale}};
	}

	# 1D piddle of length N
	elsif ( eval { $opt{scale}->isa('PDL') }
		&& $opt->{scale}->ndims == 1
		&& ($opt->{scale}->dims)[0] == $ndims
	      )
	{
	    @s = $opt{scale}->list
	}

	# everything else is an error
	else
	{
	    barf( "scale argument must be a vector of shape (n)",
		  " if theta is specified\n" );
	}

	my $t = $opt{theta};
	my $d2s = $s[1]**2 - $s[0]**2;
	my $cov = pdl( [ -$d2s * cos($t)**2 + $s[1]**2,
			  $d2s * cos($t) * sin($t) ],

		       [  $d2s * cos($t) * sin($t),
			  $d2s * cos($t)**2 + $s[0]**2] );

	# just use the rest of the code....
	$opt{scale} = $cov;
    }


    # handle scale
    # scalar -> symmetric, independent
    if ( ! ref $opt{scale} )
    {
	$cov = identity( $ndims ) * $opt{scale} ** 2;
    }

    # 1D piddle of length N
    elsif ( 'ARRAY' eq ref $opt{scale}
	    && @{$opt{scale}} == $ndims
	  )
    {
	$cov = identity( $ndims ) * pdl( $opt{scale} ) ** 2;
    }

    # 1D piddle of length N
    elsif ( eval { $opt{scale}->isa('PDL') }
	    && $opt{scale}->ndims == 1
	    && ($opt{scale}->dims)[0] == $ndims
	  )
    {
	$cov = identity( $ndims ) * $opt{scale} ** 2;
    }

    # full covariance matrix N^N piddle
    elsif ( eval { $opt{scale}->isa('PDL') }
	    && $opt{scale}->ndims == 2
	    && all( pdl( $opt{scale}->dims ) == pdl( $ndims, $ndims ) )
	  )
    {
	$cov = $opt{scale};
    }

    else
    {
	barf( "scale argument is not a scalar, an array of length $ndims,",
	      " or a piddle of shape ($ndims) or shape ($ndims,$ndims)\n" );
    }

    my $d2 = mahalanobis( $vectors, $cov,
			{ squared => 1,
			  ( defined $center ? (center => $center) : () )
			});
    my $log_pdf = -(   $ndims * log(2 * pi)
		     + log( determinant( $cov ))
		     + $d2
		   )/2;

    my $retval = $opt{log} ? $log_pdf : exp( $log_pdf );

    if ( $opt{vectors} )
    {
	$output = $retval;
    }

    else
    {
	$output .= $retval->reshape( $output->dims );
    }

    return $output;
}

=pod

=head2 mahalanobis

=for sig

  Signature: ( x(n,m), s(n,n), [o]d(m), \%options )

=for usage

  $d = mahalanobis( $v, $S, \%options );
  mahalanobis( $v, $S, $d, \%options );

=for ref

Calculate the Mahalanobis distance for one or more vectors

The Mahalanobis distance of a multivariate vector (v) from a location
(u) with a covariance matrix (S) is defined as

  dm(x,u) = sqrt( (v-u)T S^-1 (x-u) )

The input piddle representing the vectors (C<$v>) must have shape (N,m),
where C<N> is the dimension of the vector space and C<m> is the number
of vectors.

The input covariance matrix (C<$S>) must have shape (N,N).  It is I<not>
checked for validity.

=over

=item center | centre

The vector from which the distance is to be calculated.  It must have shape (N).
It defaults to the origin.

=item inverted

If true, the input matrix is the inverse of the covariance matrix.
Defaults to false.

=item squared

if true, the returned values are the distances squared.

=back


=cut

sub mahalanobis {

    # handle options.
    my $opt = 'HASH' eq ref $_[-1] ? pop( @_ ) : {};
    my %opt = PDL::Options::iparse( { center => undef,
				      inverted => 0,
				      squared  => 0,
				    }, $opt );

    my ( $x, $cov, $out ) = @_;

    # invert the matrix if it hasn't already been inverted
    $cov = $cov->inv
	unless $opt{inverted};

    my $xc;
    if ( defined $opt{center} )
    {
	my $c = PDL::Core::topdl( $opt{center} );
	$xc = $x - $c;
    }
    else
    {
	$xc = $x;
    }

    $out = zeroes(double, ($x->dims)[-1] )
      unless defined $out;

    inner2( $xc, $cov, $xc, $out );

    $out->inplace->sqrt unless $opt{squared};

    return $out;
}

*PDL::mahalanobis = \&mahalanobis;


__END__



=head1 SEE ALSO


PDL::Func.

=head1 VERSION

Version 0.01

=head1 LICENSE AND COPYRIGHT

Copyright (c) 2010 The Smithsonian Astrophysical Observatory

CXC::PDL::FuncND is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=head1 AUTHOR

Diab Jerius  E<lt>djerius@cpan.orgE<gt>


